import numpy as np

import pyfftw.builders
from pyfftw.interfaces.numpy_fft import fftn
r = np.random.randn(32, 32, 32)

# the following transform will fail if MKL FFT routines are being linked to
# instead of the FFTW ones.  (MKL doesn't support FFT of only 1
# dimension of a 3D array).
# see:  https://github.com/pyFFTW/pyFFTW/issues/40
fftn(r, axes=(0, ))



#!/usr/bin/env python
#
# Copyright 2014 Knowledge Economy Developments Ltd
#
# Henry Gomersall
# heng@kedevelopments.co.uk
#
# Michael McNeil Forbes
# michael.forbes+pyfftw@gmail.com
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

'''
Utility functions for the interfaces routines
'''

import pyfftw.builders as builders
import pyfftw
import numpy
from . import cache


def _Xfftn(a, s, axes, overwrite_input, planner_effort,
        threads, auto_align_input, auto_contiguous,
        calling_func, normalise_idft=True, ortho=False):

    work_with_copy = False

    a = numpy.asanyarray(a)

    try:
        s = tuple(s)
    except TypeError:
        pass

    try:
        axes = tuple(axes)
    except TypeError:
        pass

    if calling_func in ('irfft2', 'irfftn'):
        # overwrite_input is not an argument to irfft2 or irfftn
        args = (planner_effort, threads, auto_align_input, auto_contiguous)

        if not overwrite_input:
            # Only irfft2 and irfftn have overwriting the input
            # as the default (and so require the input array to
            # be reloaded).
            work_with_copy = True
    else:
        args = (overwrite_input, planner_effort, threads,
                auto_align_input, auto_contiguous)

        if not a.flags.writeable:
            # Special case of a locked array - always work with a
            # copy.  See issue #92.
            work_with_copy = True

            if overwrite_input:
                raise ValueError('overwrite_input cannot be True when the ' +
                                 'input array flags.writeable is False')

    if work_with_copy:
        # We make the copy before registering the key so that the
        # copy's stride information will be cached since this will be
        # used for planning.  Make sure the copy is byte aligned to
        # prevent further copying
        a_original = a
        a = pyfftw.empty_aligned(shape=a.shape, dtype=a.dtype)
        a[...] = a_original

    if cache.is_enabled():
        alignment = a.ctypes.data % pyfftw.simd_alignment

        key = (calling_func, a.shape, a.strides, a.dtype, s.__hash__(),
               axes.__hash__(), alignment, args)

        try:
            if key in cache._fftw_cache:
                FFTW_object = cache._fftw_cache.lookup(key)
            else:
                FFTW_object = None

        except KeyError:
            # This occurs if the object has fallen out of the cache between
            # the check and the lookup
            FFTW_object = None

    if not cache.is_enabled() or FFTW_object is None:

        # If we're going to create a new FFTW object and are not
        # working with a copy, then we need to copy the input array to
        # preserve it, otherwise we can't actually  take the transform
        # of the input array! (in general, we have to assume that the
        # input array will be destroyed during planning).
        if not work_with_copy:
            a_copy = a.copy()

        planner_args = (a, s, axes) + args

        FFTW_object = getattr(builders, calling_func)(*planner_args)

        # Only copy if the input array is what was actually used
        # (otherwise it shouldn't be overwritten)
        if not work_with_copy and FFTW_object.input_array is a:
            a[:] = a_copy

        if cache.is_enabled():
            cache._fftw_cache.insert(FFTW_object, key)

        output_array = FFTW_object(normalise_idft=normalise_idft, ortho=ortho)

    else:
        orig_output_array = FFTW_object.output_array
        output_shape = orig_output_array.shape
        output_dtype = orig_output_array.dtype
        output_alignment = FFTW_object.output_alignment

        output_array = pyfftw.empty_aligned(
            output_shape, output_dtype, n=output_alignment)

        FFTW_object(input_array=a, output_array=output_array,
                normalise_idft=normalise_idft, ortho=ortho)

    return output_array



#!/usr/bin/env python
#
# Copyright 2014 Knowledge Economy Developments Ltd
# Copyright 2014 David Wells
#
# Henry Gomersall
# heng@kedevelopments.co.uk
# David Wells
# drwells <at> vt.edu
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

'''
A set of utility functions for use with the builders. Users should
not need to use the functions directly, but they are included here for
completeness and to aid with understanding of what is happening behind
the scenes.

Certainly, users may encounter instances of
:class:`~pyfftw.builders._utils._FFTWWrapper`.

Everything documented in this module is *not* part of the public API
and may change in future versions.
'''

import multiprocessing
import pyfftw
import numpy
import warnings
from .. import _threading_type
from .. import config

__all__ = ['_FFTWWrapper', '_rc_dtype_pairs', '_default_dtype', '_Xfftn',
        '_setup_input_slicers', '_compute_array_shapes', '_precook_1d_args',
        '_cook_nd_args']

_valid_efforts = ('FFTW_ESTIMATE', 'FFTW_MEASURE',
        'FFTW_PATIENT', 'FFTW_EXHAUSTIVE')

# Looking up a real dtype in here returns the complex complement of the same
# precision, and vice versa.
# It is necessary to use .char as the keys due to MSVC mapping long
# double to double and the way that numpy handles this.
_rc_dtype_pairs = {}
_default_dtype = None

# Double precision is the default default precision. Prefer casting to higher precision if
# possible. If missing, long double is mapped to double and so we lose less
# precision than by converting to single.
if '64' in pyfftw._supported_types:
    _default_dtype = numpy.dtype('float64')
    _rc_dtype_pairs.update({
        numpy.dtype('float64').char: numpy.dtype('complex128'),
        numpy.dtype('complex128').char: numpy.dtype('float64')})
if 'ld' in pyfftw._supported_types:
    if _default_dtype is None:
        _default_dtype = numpy.dtype('longdouble')
    _rc_dtype_pairs.update({
         numpy.dtype('longdouble').char: numpy.dtype('clongdouble'),
         numpy.dtype('clongdouble').char: numpy.dtype('longdouble')})
if '32' in pyfftw._supported_types:
    if _default_dtype is None:
        _default_dtype = numpy.dtype('float32')
    _rc_dtype_pairs.update({
        numpy.dtype('float32').char: numpy.dtype('complex64'),
        numpy.dtype('complex64').char: numpy.dtype('float32')})
if _default_dtype is None:
    raise NotImplementedError("No default precision available")


def _default_effort(effort):
    if effort is None:
        return config.PLANNER_EFFORT
    else:
        return effort


def _default_threads(threads):
    if threads is None:
        if config.NUM_THREADS <= 0:
            return multiprocessing.cpu_count()
        return config.NUM_THREADS
    else:
        if threads > 1 and _threading_type is None:
            raise ValueError("threads > 1 requested, but pyFFTW was not built "
                             "with multithreaded FFTW.")
        elif threads <= 0:
            return multiprocessing.cpu_count()
        return threads


def _unitary(norm):
    """_unitary() utility copied from numpy"""
    if norm not in (None, "ortho"):
        raise ValueError("Invalid norm value %s, should be None or \"ortho\"."
                         % norm)
    return norm is not None


def _norm_args(norm):
    """ pass the proper normalization-related keyword arguments. """
    if _unitary(norm):
        ortho = True
        normalise_idft = False
    else:
        ortho = False
        normalise_idft = True
    return dict(normalise_idft=normalise_idft, ortho=ortho)


def _Xfftn(a, s, axes, overwrite_input,
        planner_effort, threads, auto_align_input, auto_contiguous,
        avoid_copy, inverse, real, normalise_idft=True, ortho=False):
    '''Generic transform interface for all the transforms. No
    defaults exist. The transform must be specified exactly.
    '''
    a_orig = a
    invreal = inverse and real

    if inverse:
        direction = 'FFTW_BACKWARD'
    else:
        direction = 'FFTW_FORWARD'

    if planner_effort not in _valid_efforts:
        raise ValueError('Invalid planner effort: ', planner_effort)

    s, axes = _cook_nd_args(a, s, axes, invreal)

    input_shape, output_shape = _compute_array_shapes(
            a, s, axes, inverse, real)

    a_is_complex = numpy.iscomplexobj(a)

    # Make the input dtype correct by transforming to an available type
    if a.dtype.char not in _rc_dtype_pairs:
        dtype = _default_dtype
        if a.dtype == numpy.dtype('float16') and '32' in pyfftw._supported_types:
            # convert half-precision to single precision, if available
            dtype = numpy.dtype('float32')

        # warn when losing precision but not when using a higher precision
        if dtype.itemsize < a.dtype.itemsize:
            warnings.warn("Narrowing conversion from %s to %s precision" % (a.dtype, dtype))

        if not real or inverse:
            # It's going to be complex
            dtype = numpy.dtype(_rc_dtype_pairs[dtype.char])

        # finally convert the input array
        a = numpy.asarray(a, dtype=dtype)
    elif not (real and not inverse) and not a_is_complex:
        # We need to make it a complex dtype
        a = numpy.asarray(a, dtype=_rc_dtype_pairs[a.dtype.char])

    elif (real and not inverse) and a_is_complex:
        # It should be real
        a = numpy.asarray(a, dtype=_rc_dtype_pairs[a.dtype.char])

    # Make the output dtype correct
    if not real:
        output_dtype = a.dtype

    else:
        output_dtype = _rc_dtype_pairs[a.dtype.char]

    if not avoid_copy:
        a_copy = a.copy()

    output_array = pyfftw.empty_aligned(output_shape, output_dtype)

    flags = [planner_effort]

    if not auto_align_input:
        flags.append('FFTW_UNALIGNED')

    if overwrite_input:
        flags.append('FFTW_DESTROY_INPUT')

    if not a.shape == input_shape:

        if avoid_copy:
            raise ValueError('Cannot avoid copy: '
                    'The transform shape is not the same as the array size. '
                    '(from avoid_copy flag)')

        # This means we need to use an _FFTWWrapper object
        # and so need to create slicers.
        update_input_array_slicer, FFTW_array_slicer = (
                _setup_input_slicers(a.shape, input_shape))

        # Also, the input array will be a different shape to the shape of
        # `a`, so we need to create a new array.
        input_array = pyfftw.empty_aligned(input_shape, a.dtype)

        FFTW_object = _FFTWWrapper(input_array, output_array, axes, direction,
                flags, threads, input_array_slicer=update_input_array_slicer,
                FFTW_array_slicer=FFTW_array_slicer,
                normalise_idft=normalise_idft, ortho=ortho)

        # We copy the data back into the internal FFTW object array
        internal_array = FFTW_object.input_array
        internal_array[:] = 0
        internal_array[FFTW_array_slicer] = (
                a_copy[update_input_array_slicer])

    else:
        # Otherwise we can use `a` as-is

        input_array = a

        if auto_contiguous:
            # We only need to create a new array if it's not already
            # contiguous
            if not (a.flags['C_CONTIGUOUS'] or a.flags['F_CONTIGUOUS']):
                if avoid_copy:
                    raise ValueError('Cannot avoid copy: '
                            'The input array is not contiguous and '
                            'auto_contiguous is set. (from avoid_copy flag)')

                input_array = pyfftw.empty_aligned(a.shape, a.dtype)

        if (auto_align_input and not pyfftw.is_byte_aligned(input_array)):

            if avoid_copy:
                raise ValueError('Cannot avoid copy: '
                        'The input array is not aligned and '
                        'auto_align is set. (from avoid_copy flag)')

            input_array = pyfftw.byte_align(input_array)


        FFTW_object = pyfftw.FFTW(input_array, output_array, axes, direction,
                flags, threads, normalise_idft=normalise_idft, ortho=ortho)

        if not avoid_copy:
            # Copy the data back into the (likely) destroyed array
            FFTW_object.input_array[:] = a_copy

    return FFTW_object


class _FFTWWrapper(pyfftw.FFTW):
    ''' A class that wraps :class:`pyfftw.FFTW`, providing a slicer on the input
    stage during calls to :meth:`~pyfftw.builders._utils._FFTWWrapper.__call__`.
    '''

    def __init__(self, input_array, output_array, axes=[-1],
            direction='FFTW_FORWARD', flags=['FFTW_MEASURE'],
            threads=1, input_array_slicer=None, FFTW_array_slicer=None,
            normalise_idft=True, ortho=False):
        '''The arguments are as per :class:`pyfftw.FFTW`, but with the addition
        of 2 keyword arguments: ``input_array_slicer`` and
        ``FFTW_array_slicer``.

        These arguments represent 2 slicers: ``input_array_slicer`` slices
        the input array that is passed in during a call to instances of this
        class, and ``FFTW_array_slicer`` slices the internal array.

        The arrays that are returned from both of these slicing operations
        should be the same size. The data is then copied from the sliced
        input array into the sliced internal array.
        '''

        self._input_array_slicer = input_array_slicer
        self._FFTW_array_slicer = FFTW_array_slicer
        self._normalise_idft = normalise_idft
        self._ortho = ortho

        if 'FFTW_DESTROY_INPUT' in flags:
            self._input_destroyed = True
        else:
            self._input_destroyed = False

        pyfftw.FFTW.__init__(self, input_array, output_array,
                             axes, direction, flags, threads)

    def __call__(self, input_array=None, output_array=None,
            normalise_idft=None, ortho=None):
        '''Wrap :meth:`pyfftw.FFTW.__call__` by firstly slicing the
        passed-in input array and then copying it into a sliced version
        of the internal array. These slicers are set at instantiation.

        When input array is not ``None``, this method always results in
        a copy. Consequently, the alignment and dtype are maintained in
        the internal array.

        ``output_array`` and ``normalise_idft`` are passed through to
        :meth:`pyfftw.FFTW.__call__` untouched.
        '''

        if input_array is not None:
            # Do the update here (which is a copy, so it's alignment
            # safe etc).

            internal_input_array = self.input_array
            input_array = numpy.asanyarray(input_array)

            if self._input_destroyed:
                internal_input_array[:] = 0

            sliced_internal = internal_input_array[self._FFTW_array_slicer]
            sliced_input = input_array[self._input_array_slicer]

            if sliced_internal.shape != sliced_input.shape:
                raise ValueError('Invalid input shape: '
                        'The new input array should be the same shape '
                        'as the input array used to instantiate the '
                        'object.')

            sliced_internal[:] = sliced_input

        if normalise_idft is None:
            normalise_idft = self._normalise_idft

        if ortho is None:
            ortho = self._ortho

        output = super(_FFTWWrapper, self).__call__(input_array=None,
                output_array=output_array, normalise_idft=normalise_idft,
                ortho=ortho)

        return output


def _setup_input_slicers(a_shape, input_shape):
    ''' This function returns two slicers that are to be used to
    copy the data from the input array to the FFTW object internal
    array, which can then be passed to _FFTWWrapper:

    ``(update_input_array_slicer, FFTW_array_slicer)``

    On calls to :class:`~pyfftw.builders._utils._FFTWWrapper` objects,
    the input array is copied in as:

    ``FFTW_array[FFTW_array_slicer] = input_array[update_input_array_slicer]``
    '''

    # default the slicers to include everything
    update_input_array_slicer = (
            [slice(None)]*len(a_shape))
    FFTW_array_slicer = [slice(None)]*len(a_shape)

    # iterate over each dimension and modify the slicer and FFTW dimension
    for axis in range(len(a_shape)):

        if a_shape[axis] > input_shape[axis]:
            update_input_array_slicer[axis] = (
                    slice(0, input_shape[axis]))

        elif a_shape[axis] < input_shape[axis]:
            FFTW_array_slicer[axis] = (
                    slice(0, a_shape[axis]))
            update_input_array_slicer[axis] = (
                    slice(0, a_shape[axis]))

        else:
            # If neither of these, we use the whole dimension.
            update_input_array_slicer[axis] = (
                    slice(0, a_shape[axis]))

    return tuple(update_input_array_slicer), tuple(FFTW_array_slicer)

def _compute_array_shapes(a, s, axes, inverse, real):
    '''Given a passed in array ``a``, and the rest of the arguments
    (that have been fleshed out with
    :func:`~pyfftw.builders._utils._cook_nd_args`), compute
    the shape the input and output arrays need to be in order
    to satisfy all the requirements for the transform. The input
    shape *may* be different to the shape of a.

    returns:
    ``(input_shape, output_shape)``
    '''
    # Start with the shape of a
    orig_domain_shape = list(a.shape)
    fft_domain_shape = list(a.shape)

    try:
        for n, axis in enumerate(axes):
            orig_domain_shape[axis] = s[n]
            fft_domain_shape[axis] = s[n]

        if real:
            fft_domain_shape[axes[-1]] = s[-1]//2 + 1

    except IndexError:
        raise IndexError('Invalid axes: '
                'At least one of the passed axes is invalid.')

    if inverse:
        input_shape = fft_domain_shape
        output_shape = orig_domain_shape
    else:
        input_shape = orig_domain_shape
        output_shape = fft_domain_shape

    return tuple(input_shape), tuple(output_shape)

def _precook_1d_args(a, n, axis):
    '''Turn ``*(n, axis)`` into ``(s, axes)``
    '''
    if n is not None:
        s = [int(n)]
    else:
        s = None

    # Force an error with an invalid axis
    a.shape[axis]

    return s, (axis,)

def _cook_nd_args(a, s=None, axes=None, invreal=False):
    '''Similar to :func:`numpy.fft.fftpack._cook_nd_args`.
    '''

    if axes is None:
        if s is None:
            len_s = len(a.shape)
        else:
            len_s = len(s)

        axes = list(range(-len_s, 0))

    if s is None:
        s = list(numpy.take(a.shape, axes))

        if invreal:
            s[-1] = (a.shape[axes[-1]] - 1) * 2


    if len(s) != len(axes):
        raise ValueError('Shape error: '
                'Shape and axes have different lengths.')

    if len(s) > len(a.shape):
        raise ValueError('Shape error: '
                'The length of s or axes cannot exceed the dimensionality '
                'of the input array, a.')

    return tuple(s), tuple(axes)



#!/usr/bin/env python
#
# Henry Gomersall
# heng@kedevelopments.co.uk
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

'''
This module implements those functions that replace aspects of the
:mod:`dask.fft` module. This module *provides* the entire documented
namespace of :mod:`dask.fft`, but those functions that are not included
here are imported directly from :mod:`dask.fft`.


It is notable that unlike :mod:`numpy.fftpack`, which :mod:`dask.fft`
wraps, these functions will generally return an output array with the
same precision as the input array, and the transform that is chosen is
chosen based on the precision of the input array. That is, if the input
array is 32-bit floating point, then the transform will be 32-bit floating
point and so will the returned array. Half precision input will be
converted to single precision.  Otherwise, if any type conversion is
required, the default will be double precision.

The exceptions raised by each of these functions are mostly as per their
equivalents in :mod:`dask.fft`, though there are some corner cases in
which this may not be true.
'''

from . import numpy_fft as _numpy_fft
from dask.array.fft import (
    fft_wrap,
    fftfreq,
    rfftfreq,
    fftshift,
    ifftshift,
)

fft = fft_wrap(_numpy_fft.fft)
fft2 = fft_wrap(_numpy_fft.fft2)
fftn = fft_wrap(_numpy_fft.fftn)
ifft = fft_wrap(_numpy_fft.ifft)
ifft2 = fft_wrap(_numpy_fft.ifft2)
ifftn = fft_wrap(_numpy_fft.ifftn)
rfft = fft_wrap(_numpy_fft.rfft)
rfft2 = fft_wrap(_numpy_fft.rfft2)
rfftn = fft_wrap(_numpy_fft.rfftn)
irfft = fft_wrap(_numpy_fft.irfft)
irfft2 = fft_wrap(_numpy_fft.irfft2)
irfftn = fft_wrap(_numpy_fft.irfftn)
hfft = fft_wrap(_numpy_fft.hfft)
ihfft = fft_wrap(_numpy_fft.ihfft)



"""Global configuration variables for PyFFTW.

The approach taken here was adapated from Numba's config.py.
"""
from __future__ import print_function, division, absolute_import

import os
import multiprocessing
import warnings

from .pyfftw import _threading_type


class _EnvReloader(object):

    def __init__(self):
        self.reset()

    def reset(self):
        self.old_environ = {}
        self.update(force=True)

    def update(self, force=False):
        new_environ = {}

        # read local env var OMP_NUM_THREADS and any starting with PYFFTW_
        for name, value in os.environ.items():
            if name.startswith('PYFFTW_') or name == 'OMP_NUM_THREADS':
                new_environ[name] = value
        # We update the config variables if at least one PYFFTW environment
        # variable was modified.  This lets the user modify values
        # directly in the config module without having them when
        # reload_config() is called by the compiler.
        if force or self.old_environ != new_environ:
            self.process_environ(new_environ)
            # Store a copy
            self.old_environ = dict(new_environ)

    def process_environ(self, environ):
        def _readenv(name, ctor, default):
            value = environ.get(name)
            if value is None:
                return default() if callable(default) else default
            try:
                return ctor(value)
            except Exception:
                warnings.warn("environ %s defined but failed to parse '%s'" %
                              (name, value), RuntimeWarning)
                return default

        def optional_str(x):
            return str(x) if x is not None else None

        if _threading_type is None:
            NUM_THREADS = 1
        else:
            if (_threading_type == "OMP" and
                    "PYFFTW_NUM_THREADS" not in environ):
                # fallback to OMP_NUM_THREADS if PYFFTW_NUM_THREADS undefined
                NUM_THREADS = _readenv("OMP_NUM_THREADS", int, 1)
            else:
                NUM_THREADS = _readenv("PYFFTW_NUM_THREADS", int, 1)
            # if user requested <= 0 threads, use the maximum available
            if NUM_THREADS <= 0:
                NUM_THREADS = multiprocessing.cpu_count()

        PLANNER_EFFORT = _readenv(
            "PYFFTW_PLANNER_EFFORT", str, "FFTW_ESTIMATE")

        # Inject the configuration values into the module globals
        for name, value in locals().copy().items():
            if name.isupper():
                globals()[name] = value

_env_reloader = _EnvReloader()


def _reload_config():
    """
    Reload the configuration from environment variables, if necessary.
    """
    _env_reloader.update()



#!/usr/bin/env python
#
# Copyright 2014 Knowledge Economy Developments Ltd
# Copyright 2014 David Wells
#
# Henry Gomersall
# heng@kedevelopments.co.uk
# David Wells
# drwells <at> vt.edu
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

'''
Overview
""""""""

This module contains a set of functions that return
:class:`pyfftw.FFTW` objects.

The interface to create these objects is mostly the same as
:mod:`numpy.fft`, only instead of the call returning the result of the
FFT, a :class:`pyfftw.FFTW` object is returned that performs that FFT
operation when it is called. Users should be familiar with
:mod:`!numpy.fft` before reading on.

In the case where the shape argument, ``s`` (or ``n`` in the
1-dimensional case), dictates that the passed-in input array be copied
into a different processing array, the returned object is an
instance of a child class of :class:`pyfftw.FFTW`,
:class:`~pyfftw.builders._utils._FFTWWrapper`, which wraps the call
method in order to correctly perform that copying. That is, subsequent
calls to the object (i.e. through
:meth:`~pyfftw.builders._utils._FFTWWrapper.__call__`) should occur
with an input array that can be sliced to the same size as the
expected internal array. Note that a side effect of this is that
subsequent calls to the object can be made with an array that is
*bigger* than the original (but not smaller).

Only the call method is wrapped; :meth:`~pyfftw.FFTW.update_arrays`
still expects an array with the correct size, alignment, dtype etc for
the :class:`pyfftw.FFTW` object.

When the internal input array is bigger along any axis than the input
array that is passed in (due to ``s`` dictating a larger size), then the
extra entries are padded with zeros. This is a one time action. If the
internal input array is then extracted using
:attr:`pyfftw.FFTW.input_array`, it is possible to
persistently fill the padding space with whatever the user desires, so
subsequent calls with a new input only overwrite the values that aren't
padding (even if the array that is used for the call is bigger than the
original - see the point above about bigger arrays being sliced to
fit).

The precision of the FFT operation is acquired from the input array.
If an array is passed in that is not of float type, or is of an
unknown float type, an attempt is made to convert the array to a
double precision array. This results in a copy being made.

If an array of the incorrect complexity is passed in (e.g. a complex
array is passed to a real transform routine, or vice-versa), then an
attempt is made to convert the array to an array of the correct
complexity. This results in a copy being made.

Although the array that is internal to the :class:`pyfftw.FFTW` object
will be correctly loaded with the values within the input array, it is
not necessarily the case that the internal array *is* the input array.
The actual internal input array can always be retrieved with
:attr:`pyfftw.FFTW.input_array`.

The behaviour of the ``norm`` option in all builder routines matches that of
the corresponding numpy functions.  In short, if ``norm`` is ``None``, then the
output from the forward DFT is unscaled and the inverse DFT is scaled by 1/N,
where N is the product of the lengths of input array on which the FFT is taken.
If ``norm == 'ortho'``, then the output of both forward and inverse DFT
operations are scaled by 1/sqrt(N).

**Example:**

.. doctest::

    >>> import pyfftw
    >>> a = pyfftw.empty_aligned(4, dtype='complex128')
    >>> fft = pyfftw.builders.fft(a)
    >>> a[:] = [1, 2, 3, 4]
    >>> fft() # returns the output
    array([ 10.+0.j,  -2.+2.j,  -2.+0.j,  -2.-2.j])

More examples can be found in the :doc:`tutorial </source/tutorial>`.

Supported Functions and Caveats
"""""""""""""""""""""""""""""""

The following functions are supported. They can be used with the
same calling signature as their respective functions in
:mod:`numpy.fft`.

**Standard FFTs**

* :func:`~pyfftw.builders.fft`
* :func:`~pyfftw.builders.ifft`
* :func:`~pyfftw.builders.fft2`
* :func:`~pyfftw.builders.ifft2`
* :func:`~pyfftw.builders.fftn`
* :func:`~pyfftw.builders.ifftn`

**Real FFTs**

* :func:`~pyfftw.builders.rfft`
* :func:`~pyfftw.builders.irfft`
* :func:`~pyfftw.builders.rfft2`
* :func:`~pyfftw.builders.irfft2`
* :func:`~pyfftw.builders.rfftn`
* :func:`~pyfftw.builders.irfftn`

The first caveat is that the dtype of the input array must match the
transform. For example, for ``fft`` and ``ifft``, the dtype must
be complex, for ``rfft`` it must be real, and so on. The other point
to note from this is that the precision of the transform matches the
precision of the input array. So, if a single precision input array is
passed in, then a single precision transform will be used.

The second caveat is that repeated axes are handled differently; with
the returned :class:`pyfftw.FFTW` object, axes that are repeated in the
axes argument are considered only once, as compared to :mod:`numpy.fft`
in which repeated axes results in the DFT being taken along that axes
as many times as the axis occurs (this is down to the underlying
library).

Note that unless the ``auto_align_input`` argument to the function
is set to ``True``, the ``'FFTW_UNALIGNED'`` :ref:`flag <FFTW_flags>`
is set in the returned :class:`pyfftw.FFTW` object. This disables some
of the FFTW optimisations that rely on aligned arrays. Also worth
noting is that the ``auto_align_input`` flag only results in a copy
when calling the resultant :class:`pyfftw.FFTW` object if the input
array is not already aligned correctly.

.. _builders_args:

Additional Arguments
""""""""""""""""""""

In addition to the arguments that are present with their complementary
functions in :mod:`numpy.fft`, each of these functions also offers the
following additional keyword arguments:

* ``overwrite_input``: Whether or not the input array can be
  overwritten during the transform. This sometimes results in a faster
  algorithm being made available. It causes the ``'FFTW_DESTROY_INPUT'``
  flag to be passed to the :class:`pyfftw.FFTW` object. This flag is not
  offered for the multi-dimensional inverse real transforms, as FFTW is
  unable to not overwrite the input in that case.

* ``planner_effort``: A string dictating how much effort is spent
  in planning the FFTW routines. This is passed to the creation
  of the :class:`pyfftw.FFTW` object as an entry in the flags list.
  They correspond to flags passed to the :class:`pyfftw.FFTW` object.

  The valid strings, in order of their increasing impact on the time
  to compute  are:
  ``'FFTW_ESTIMATE'``, ``config.PLANNER_EFFORT`` (default), ``'FFTW_PATIENT'``
  and ``'FFTW_EXHAUSTIVE'``.

  The `Wisdom
  <http://www.fftw.org/fftw3_doc/Words-of-Wisdom_002dSaving-Plans.html>`_
  that FFTW has accumulated or has loaded (through
  :func:`pyfftw.import_wisdom`) is used during the creation of
  :class:`pyfftw.FFTW` objects.

* ``threads``: The number of threads used to perform the FFT.

* ``auto_align_input``: Correctly byte align the input array for optimal
  usage of vector instructions. This can lead to a substantial speedup.

  Setting this argument to ``True`` makes sure that the input array
  is correctly aligned. It is possible to correctly byte align the array
  prior to calling this function (using, for example,
  :func:`pyfftw.byte_align`). If and only if a realignment is
  necessary is a new array created. If a new array *is* created, it is
  up to the calling code to acquire that new input array using
  :attr:`pyfftw.FFTW.input_array`.

  The resultant :class:`pyfftw.FFTW` object that is created will be
  designed to operate on arrays that are aligned. If the object is
  called with an unaligned array, this would result in a copy. Despite
  this, it may still be faster to set the ``auto_align_input`` flag
  and incur a copy with unaligned arrays than to set up an object
  that uses aligned arrays.

  It's worth noting that just being aligned may not be sufficient to
  create the fastest possible transform. For example, if the array is not
  contiguous (i.e. certain axes have gaps in memory between slices), it
  may be faster to plan a transform for a contiguous array, and then rely
  on the array being copied in before the transform (which
  :class:`pyfftw.FFTW` will handle for you). The ``auto_contiguous``
  argument controls whether this function also takes care of making sure
  the array is contiguous or not.

* ``auto_contiguous``: Make sure the input array is contiguous in
  memory before performing the transform on it. If the array is not
  contiguous, it is copied into an interim array. This is because it
  is often faster to copy the data before the transform and then transform
  a contiguous array than it is to try to take the transform of a
  non-contiguous array. This is particularly true in conjunction with
  the ``auto_align_input`` argument which is used to make sure that the
  transform is taken of an aligned array.

  Like ``auto_align_input``, If a new array is created, it is
  up to the calling code to acquire that new input array using
  :attr:`pyfftw.FFTW.input_array`.

* ``avoid_copy``: By default, these functions will always create a copy
  (and sometimes more than one) of the passed in input array. This is
  because the creation of the :class:`pyfftw.FFTW` object generally
  destroys the contents of the input array. Setting this argument to
  ``True`` will try not to create a copy of the input array, likely
  resulting in the input array being destroyed. If it is not possible
  to create the object without a copy being made, a ``ValueError`` is
  raised.

  Example situations that require a copy, and so cause the exception
  to be raised when this flag is set:

  * The shape of the FFT input as dictated by ``s`` is
    necessarily different from the shape of the passed-in array.

  * The dtypes are incompatible with the FFT routine.

  * The ``auto_contiguous`` or ``auto_align`` flags are True and
    the input array is not already contiguous or aligned.

  This argument is distinct from ``overwrite_input`` in that it only
  influences a copy during the creation of the object. It changes no
  flags in the :class:`pyfftw.FFTW` object.

The exceptions raised by each of these functions are as per their
equivalents in :mod:`numpy.fft`, or as documented above.
'''

from ._utils import (_precook_1d_args, _Xfftn, _norm_args, _default_effort,
                     _default_threads)

__all__ = ['fft','ifft', 'fft2', 'ifft2', 'fftn',
           'ifftn', 'rfft', 'irfft', 'rfft2', 'irfft2', 'rfftn',
           'irfftn']


def fft(a, n=None, axis=-1, overwrite_input=False,
        planner_effort=None, threads=None,
        auto_align_input=True, auto_contiguous=True,
        avoid_copy=False, norm=None):
    '''Return a :class:`pyfftw.FFTW` object representing a 1D FFT.

    The first three arguments are as per :func:`numpy.fft.fft`;
    the rest of the arguments are documented
    :ref:`in the module docs <builders_args>`.
    '''
    inverse = False
    real = False

    s, axes = _precook_1d_args(a, n, axis)
    planner_effort = _default_effort(planner_effort)
    threads = _default_threads(threads)

    return _Xfftn(a, s, axes, overwrite_input, planner_effort,
            threads, auto_align_input, auto_contiguous,
            avoid_copy, inverse, real, **_norm_args(norm))

def ifft(a, n=None, axis=-1, overwrite_input=False,
        planner_effort=None, threads=None,
        auto_align_input=True, auto_contiguous=True,
        avoid_copy=False, norm=None):
    '''Return a :class:`pyfftw.FFTW` object representing a 1D
    inverse FFT.

    The first three arguments are as per :func:`numpy.fft.ifft`;
    the rest of the arguments are documented
    :ref:`in the module docs <builders_args>`.
    '''

    inverse = True
    real = False

    s, axes = _precook_1d_args(a, n, axis)
    planner_effort = _default_effort(planner_effort)
    threads = _default_threads(threads)

    return _Xfftn(a, s, axes, overwrite_input, planner_effort,
            threads, auto_align_input, auto_contiguous,
            avoid_copy, inverse, real, **_norm_args(norm))


def fft2(a, s=None, axes=(-2,-1), overwrite_input=False,
        planner_effort=None, threads=None,
        auto_align_input=True, auto_contiguous=True,
        avoid_copy=False, norm=None):
    '''Return a :class:`pyfftw.FFTW` object representing a 2D FFT.

    The first three arguments are as per :func:`numpy.fft.fft2`;
    the rest of the arguments are documented
    :ref:`in the module docs <builders_args>`.
    '''


    inverse = False
    real = False
    planner_effort = _default_effort(planner_effort)
    threads = _default_threads(threads)

    return _Xfftn(a, s, axes, overwrite_input, planner_effort,
            threads, auto_align_input, auto_contiguous,
            avoid_copy, inverse, real, **_norm_args(norm))

def ifft2(a, s=None, axes=(-2,-1), overwrite_input=False,
        planner_effort=None, threads=None,
        auto_align_input=True, auto_contiguous=True,
        avoid_copy=False, norm=None):
    '''Return a :class:`pyfftw.FFTW` object representing a
    2D inverse FFT.

    The first three arguments are as per :func:`numpy.fft.ifft2`;
    the rest of the arguments are documented
    :ref:`in the module docs <builders_args>`.
    '''


    inverse = True
    real = False
    planner_effort = _default_effort(planner_effort)
    threads = _default_threads(threads)

    return _Xfftn(a, s, axes, overwrite_input, planner_effort,
            threads, auto_align_input, auto_contiguous,
            avoid_copy, inverse, real, **_norm_args(norm))


def fftn(a, s=None, axes=None, overwrite_input=False,
        planner_effort=None, threads=None,
        auto_align_input=True, auto_contiguous=True,
        avoid_copy=False, norm=None):
    '''Return a :class:`pyfftw.FFTW` object representing a n-D FFT.

    The first three arguments are as per :func:`numpy.fft.fftn`;
    the rest of the arguments are documented
    :ref:`in the module docs <builders_args>`.
    '''


    inverse = False
    real = False
    planner_effort = _default_effort(planner_effort)
    threads = _default_threads(threads)

    return _Xfftn(a, s, axes, overwrite_input, planner_effort,
            threads, auto_align_input, auto_contiguous,
            avoid_copy, inverse, real, **_norm_args(norm))

def ifftn(a, s=None, axes=None, overwrite_input=False,
        planner_effort=None, threads=None,
        auto_align_input=True, auto_contiguous=True,
        avoid_copy=False, norm=None):
    '''Return a :class:`pyfftw.FFTW` object representing an n-D
    inverse FFT.

    The first three arguments are as per :func:`numpy.fft.ifftn`;
    the rest of the arguments are documented
    :ref:`in the module docs <builders_args>`.
    '''


    inverse = True
    real = False
    planner_effort = _default_effort(planner_effort)
    threads = _default_threads(threads)

    return _Xfftn(a, s, axes, overwrite_input, planner_effort,
            threads, auto_align_input, auto_contiguous,
            avoid_copy, inverse, real, **_norm_args(norm))

def rfft(a, n=None, axis=-1, overwrite_input=False,
        planner_effort=None, threads=None,
        auto_align_input=True, auto_contiguous=True,
        avoid_copy=False, norm=None):
    '''Return a :class:`pyfftw.FFTW` object representing a 1D
    real FFT.

    The first three arguments are as per :func:`numpy.fft.rfft`;
    the rest of the arguments are documented
    :ref:`in the module docs <builders_args>`.
    '''


    inverse = False
    real = True

    s, axes = _precook_1d_args(a, n, axis)
    planner_effort = _default_effort(planner_effort)
    threads = _default_threads(threads)

    return _Xfftn(a, s, axes, overwrite_input, planner_effort,
            threads, auto_align_input, auto_contiguous,
            avoid_copy, inverse, real, **_norm_args(norm))

def irfft(a, n=None, axis=-1, overwrite_input=False,
        planner_effort=None, threads=None,
        auto_align_input=True, auto_contiguous=True,
        avoid_copy=False, norm=None):
    '''Return a :class:`pyfftw.FFTW` object representing a 1D
    real inverse FFT.

    The first three arguments are as per :func:`numpy.fft.irfft`;
    the rest of the arguments are documented
    :ref:`in the module docs <builders_args>`.
    '''


    inverse = True
    real = True

    s, axes = _precook_1d_args(a, n, axis)
    planner_effort = _default_effort(planner_effort)
    threads = _default_threads(threads)

    return _Xfftn(a, s, axes, overwrite_input, planner_effort,
            threads, auto_align_input, auto_contiguous,
            avoid_copy, inverse, real, **_norm_args(norm))

def rfft2(a, s=None, axes=(-2,-1), overwrite_input=False,
        planner_effort=None, threads=None,
        auto_align_input=True, auto_contiguous=True,
        avoid_copy=False, norm=None):
    '''Return a :class:`pyfftw.FFTW` object representing a 2D
    real FFT.

    The first three arguments are as per :func:`numpy.fft.rfft2`;
    the rest of the arguments are documented
    :ref:`in the module docs <builders_args>`.
    '''

    inverse = False
    real = True
    planner_effort = _default_effort(planner_effort)
    threads = _default_threads(threads)

    return _Xfftn(a, s, axes, overwrite_input, planner_effort,
            threads, auto_align_input, auto_contiguous,
            avoid_copy, inverse, real, **_norm_args(norm))

def irfft2(a, s=None, axes=(-2,-1),
        planner_effort=None, threads=None,
        auto_align_input=True, auto_contiguous=True,
        avoid_copy=False, norm=None):
    '''Return a :class:`pyfftw.FFTW` object representing a 2D
    real inverse FFT.

    The first three arguments are as per :func:`numpy.fft.irfft2`;
    the rest of the arguments are documented
    :ref:`in the module docs <builders_args>`.
    '''


    inverse = True
    real = True

    overwrite_input = True
    planner_effort = _default_effort(planner_effort)
    threads = _default_threads(threads)

    return _Xfftn(a, s, axes, overwrite_input, planner_effort,
            threads, auto_align_input, auto_contiguous,
            avoid_copy, inverse, real, **_norm_args(norm))


def rfftn(a, s=None, axes=None, overwrite_input=False,
        planner_effort=None, threads=None,
        auto_align_input=True, auto_contiguous=True,
        avoid_copy=False, norm=None):
    '''Return a :class:`pyfftw.FFTW` object representing an n-D
    real FFT.

    The first three arguments are as per :func:`numpy.fft.rfftn`;
    the rest of the arguments are documented
    :ref:`in the module docs <builders_args>`.
    '''


    inverse = False
    real = True
    planner_effort = _default_effort(planner_effort)
    threads = _default_threads(threads)

    return _Xfftn(a, s, axes, overwrite_input, planner_effort,
            threads, auto_align_input, auto_contiguous,
            avoid_copy, inverse, real, **_norm_args(norm))


def irfftn(a, s=None, axes=None,
        planner_effort=None, threads=None,
        auto_align_input=True, auto_contiguous=True,
        avoid_copy=False, norm=None):
    '''Return a :class:`pyfftw.FFTW` object representing an n-D
    real inverse FFT.

    The first three arguments are as per :func:`numpy.fft.rfftn`;
    the rest of the arguments are documented
    :ref:`in the module docs <builders_args>`.
    '''


    inverse = True
    real = True

    overwrite_input = True
    planner_effort = _default_effort(planner_effort)
    threads = _default_threads(threads)

    return _Xfftn(a, s, axes, overwrite_input, planner_effort,
            threads, auto_align_input, auto_contiguous,
            avoid_copy, inverse, real, **_norm_args(norm))



#!/usr/bin/env python
#
# Copyright 2015 Knowledge Economy Developments Ltd
#
# Henry Gomersall
# heng@kedevelopments.co.uk
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

'''
During calls to functions implemented in :mod:`pyfftw.interfaces`, a
:class:`pyfftw.FFTW` object is necessarily created. Although the time to
create a new :class:`pyfftw.FFTW` is short (assuming that the planner
possesses the necessary wisdom to create the plan immediately), it may
still take longer than a short transform.

This module implements a method by which objects that are created through
:mod:`pyfftw.interfaces` are temporarily cached. If an equivalent
transform is then performed within a short period, the object is acquired
from the cache rather than a new one created. The equivalency is quite
conservative and in practice means that if any of the arguments change, or
if the properties of the array (shape, strides, dtype) change in any way, then
the cache lookup will fail.

The cache temporarily stores a copy of any interim :class:`pyfftw.FFTW`
objects that are created. If they are not used for some period of time,
which can be set with :func:`pyfftw.interfaces.cache.set_keepalive_time`,
then they are removed from the cache (liberating any associated memory).
The default keepalive time is 0.1 seconds.

Enable the cache by calling :func:`pyfftw.interfaces.cache.enable`.
Disable it by calling :func:`pyfftw.interfaces.cache.disable`. By default,
the cache is disabled.

Note that even with the cache enabled, there is a fixed overhead associated
with lookups. This means that for small transforms, the overhead may exceed
the transform. At this point, it's worth looking at using :class:`pyfftw.FFTW`
directly.

When the cache is enabled, the module spawns a new thread to keep track
of the objects. If :mod:`threading` is not available, then the cache
is not available and trying to use it will raise an ImportError exception.

The actual implementation of the cache is liable to change, but the
documented API is stable.
'''

try:
    import threading as _threading
    _threading_import_error = None
except ImportError as e:
    _threading_import_error = e
    _threading == None

import time
import weakref

__all__ = ['enable', 'disable', 'set_keepalive_time']

_fftw_cache = None

class CacheError(Exception):
    pass

def enable():
    '''Enable the cache.
    '''
    global _fftw_cache

    if _threading is not None:
        if _fftw_cache is None:
            _fftw_cache = _Cache()
    else:
        raise ImportError(_threading_import_error)

def disable():
    '''Disable the cache.
    '''
    global _fftw_cache
    _fftw_cache = None

def is_enabled():
    '''Return whether the cache is currently enabled.
    '''
    if _fftw_cache is None:
        return False
    else:
        return True

def set_keepalive_time(keepalive_time):
    '''Set the minimum time in seconds for which any :mod:`pyfftw.FFTW` object
    in the cache is kept alive.

    When the cache is enabled, the interim objects that are used through
    a :mod:`pyfftw.interfaces` function are cached for the time set through
    this function. If the object is not used for the that time, it is
    removed from the cache. Using the object zeros the timer.

    The time is not precise, and sets a minimum time to be alive. In
    practice, it may be quite a bit longer before the object is
    deleted from the cache (due to implementational details - e.g. contention
    from other threads).
    '''
    global _fftw_cache

    if _fftw_cache is None:
        raise CacheError('Cache is not currently enabled')
    else:
        _fftw_cache.set_keepalive_time(keepalive_time)

class _Cache(object):

    @property
    def keepalive_time(self):
        return self._keepalive_time

    def __init__(self, keepalive_time=0.1):

        self._cache_dict = {}
        self.set_keepalive_time(keepalive_time)

        # A set of objects to be kept alive during the next cull
        self._keepalive_set = set()

        self._cull_lock = _threading.Lock()
        self._keepalive_set_lock = _threading.Lock()

        self.initialised = _threading.Event()

        self._parent_thread = _threading.current_thread()
        self._close_thread_now = _threading.Event()

        self._initialised = _threading.Event()
        self._initialised.clear() # Explicitly clear it for clarity

        self._thread_object = _threading.Thread(target=_Cache._run,
                args=(weakref.proxy(self), ), name='PyFFTWCacheThread')

        self._thread_object.daemon = True
        self._thread_object.start()

        while not self._initialised.is_set():
            # This loop is necessary to stop the main thread doing
            # anything until the exception handler in _run can deal with
            # the object being deleted.
            pass

    def __del__(self):
        # Wait until the thread object has quit before
        # exiting (which it will because a reference error will
        # be raised).
        try:
            self._close_thread_now.set()

        except TypeError:
            # Not sure what's going on here, but IPython baulks on exit
            pass

    def __contains__(self, key):
        return key in self._cache_dict

    def _run(self):

        last_cull_time = time.time()

        try:
            self._initialised.set()

            while True:
                if (not self._parent_thread.is_alive() or
                    self._close_thread_now.is_set()):
                    break

                if time.time() - last_cull_time > self._keepalive_time:
                    # Perform a cull
                    last_cull_time = time.time()

                    with self._cull_lock:
                        # Operate on a copy of the cache dict
                        # so lookups continue.
                        new_cache_dict = self._cache_dict.copy()

                        with self._keepalive_set_lock:
                            # Work out which should be culled
                            cull_set = set(new_cache_dict).difference(
                                    self._keepalive_set)

                            self._keepalive_set = set()

                        for each_key in cull_set:
                            del new_cache_dict[each_key]

                        # Necessarily atomic, so no problem with
                        # the lookups continuing
                        self._cache_dict = new_cache_dict

                time.sleep(self._wakeup_time)

        except ReferenceError:
            pass

    def set_keepalive_time(self, keepalive_time=0.1):
        '''Set the minimum time in seconds for which any object in the cache
        is kept alive.

        The time is not precise, and sets a minimum time to be alive. In
        practice, it may be up to twice as long before the object is
        deleted from the cache (due to implementational details).
        '''
        self._keepalive_time = float(keepalive_time)

        if self._keepalive_time/2 > 0.1:
            self._wakeup_time = 0.1
        else:
            self._wakeup_time = self._keepalive_time/2

    def _refresh(self, key):
        '''Refresh the object referenced by key to stop it being culled
        on the next round.
        '''
        with self._keepalive_set_lock:
            self._keepalive_set.add(key)

    def insert(self, obj, key):
        '''Insert the passed object into the cache, referenced by key,
        a hashable.
        '''
        with self._cull_lock:
            self._cache_dict[key] = obj
            self._refresh(key)

    def lookup(self, key):
        '''Lookup the object referenced by key and return it, refreshing
        the cache at the same time.
        '''
        self._refresh(key)
        return self._cache_dict[key]
